from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable
from typing import Protocol


class ITimer(Protocol):
    @abstractmethod
    def Create(self, __arg0: float, __arg1: Callable[[], None]) -> ITimer: ...

    @abstractmethod
    def cancel(self) -> None: ...

    @property
    @abstractmethod
    def daemon(self) -> bool: ...

    @daemon.setter
    @abstractmethod
    def daemon(self, __arg0: bool) -> None: ...

    @abstractmethod
    def start(self) -> None: ...



from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import (Any, Generic, TypeVar)
from ..fable_library.guid import new_guid
from ..fable_library.list import (FSharpList, empty as empty_1, append, singleton, concat, map as map_1, is_empty, iterate)
from ..fable_library.map_util import (try_get_value, add_to_dict)
from ..fable_library.option import (some, to_nullable)
from ..fable_library.reflection import (TypeInfo, lambda_type, list_type, float64_type, option_type, bool_type, string_type, record_type)
from ..fable_library.seq import (indexed, find)
from ..fable_library.types import (FSharpRef, Record)
from ..fable_library.util import (curry2, ignore, get_enumerator, IEnumerable_1, equals)
from .Cells.fs_cell import (FsCell, FsCell_reflection)
from .Cells.fs_cells_collection import FsCellsCollection
from .fs_address import FsAddress
from .fs_row import FsRow
from .fs_workbook import FsWorkbook
from .fs_worksheet import FsWorksheet
from .Ranges.fs_range_address import FsRangeAddress
from .Tables.fs_table import FsTable
from .Tables.fs_table_field import FsTableField

_V = TypeVar("_V")

_T = TypeVar("_T")

def Dictionary_tryGetValue(k: Any, dict_1: Any) -> Any | None:
    pattern_input: tuple[bool, _V]
    out_arg: _V = None
    def _arrow273(__unit: None=None, k: Any=k, dict_1: Any=dict_1) -> _V:
        return out_arg

    def _arrow274(v: _V | None=None, k: Any=k, dict_1: Any=dict_1) -> None:
        nonlocal out_arg
        out_arg = v

    pattern_input = (try_get_value(dict_1, k, FSharpRef(_arrow273, _arrow274)), out_arg)
    if pattern_input[0]:
        return some(pattern_input[1])

    else: 
        return None



def Dictionary_length(dict_1: Any) -> int:
    return len(dict_1)


def _expr275(gen0: TypeInfo) -> TypeInfo:
    return record_type("FsSpreadsheet.SheetBuilder.FieldMap`1", [gen0], FieldMap_1, lambda: [("CellTransformers", list_type(lambda_type(gen0, lambda_type(FsCell_reflection(), FsCell_reflection())))), ("HeaderTransformers", list_type(lambda_type(gen0, lambda_type(FsCell_reflection(), FsCell_reflection())))), ("ColumnWidth", option_type(float64_type)), ("RowHeight", option_type(lambda_type(gen0, option_type(float64_type)))), ("AdjustToContents", bool_type), ("Hash", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class FieldMap_1(Record, Generic[_T]):
    CellTransformers: FSharpList[Callable[[_T, FsCell], FsCell]]
    HeaderTransformers: FSharpList[Callable[[_T, FsCell], FsCell]]
    ColumnWidth: float | None
    RowHeight: Callable[[_T], float | None] | None
    AdjustToContents: bool
    Hash: str

FieldMap_1_reflection = _expr275

def FieldMap_1_empty(__unit: None=None) -> FieldMap_1[Any]:
    def _arrow276(__unit: None=None) -> str:
        copy_of_struct: str = new_guid()
        return str(copy_of_struct)

    return FieldMap_1(empty_1(), empty_1(), None, None, False, _arrow276())


def FieldMap_1_create_Z3BCCB7EB(map_row: Callable[[_T, FsCell], FsCell]) -> FieldMap_1[Any]:
    empty: FieldMap_1[_T] = FieldMap_1_empty()
    return FieldMap_1(append(empty.CellTransformers, singleton(curry2(map_row))), empty.HeaderTransformers, empty.ColumnWidth, empty.RowHeight, empty.AdjustToContents, empty.Hash)


def FieldMap_1__header_Z721C83C5(self_1: FieldMap_1[Any], name: str) -> FieldMap_1[Any]:
    def transformer(_arg: _T | None=None, self_1: Any=self_1, name: Any=name) -> Callable[[FsCell], FsCell]:
        def _arrow277(cell: FsCell, _arg: Any=_arg) -> FsCell:
            cell.SetValueAs(name)
            return cell

        return _arrow277

    return FieldMap_1(self_1.CellTransformers, append(self_1.HeaderTransformers, singleton(transformer)), self_1.ColumnWidth, self_1.RowHeight, self_1.AdjustToContents, self_1.Hash)


def FieldMap_1__header_54F19B58(self_1: FieldMap_1[Any], map_header: Callable[[_T], str]) -> FieldMap_1[Any]:
    def transformer(value: _T | None=None, self_1: Any=self_1, map_header: Any=map_header) -> Callable[[FsCell], FsCell]:
        def _arrow278(cell: FsCell, value: Any=value) -> FsCell:
            cell.SetValueAs(map_header(value))
            return cell

        return _arrow278

    return FieldMap_1(self_1.CellTransformers, append(self_1.HeaderTransformers, singleton(transformer)), self_1.ColumnWidth, self_1.RowHeight, self_1.AdjustToContents, self_1.Hash)


def FieldMap_1__adjustToContents(self_1: FieldMap_1[Any]) -> FieldMap_1[Any]:
    return FieldMap_1(self_1.CellTransformers, self_1.HeaderTransformers, self_1.ColumnWidth, self_1.RowHeight, True, self_1.Hash)


def FieldMap_1_field_Z5C94BCA1(map: Callable[[_T], int]) -> FieldMap_1[Any]:
    def _arrow279(row: Any, cell: FsCell, map: Any=map) -> FsCell:
        cell.SetValueAs(map(row))
        return cell

    return FieldMap_1_create_Z3BCCB7EB(_arrow279)


def FieldMap_1_field_54F19B58(map: Callable[[_T], str]) -> FieldMap_1[Any]:
    def _arrow280(row: Any, cell: FsCell, map: Any=map) -> FsCell:
        cell.SetValueAs(map(row))
        return cell

    return FieldMap_1_create_Z3BCCB7EB(_arrow280)


def FieldMap_1_field_477D79EC(map: Callable[[_T], Any]) -> FieldMap_1[Any]:
    def _arrow281(row: Any, cell: FsCell, map: Any=map) -> FsCell:
        cell.SetValueAs(map(row))
        return cell

    return FieldMap_1_create_Z3BCCB7EB(_arrow281)


def FieldMap_1_field_Z1D55A0D7(map: Callable[[_T], bool]) -> FieldMap_1[Any]:
    def _arrow282(row: Any, cell: FsCell, map: Any=map) -> FsCell:
        cell.SetValueAs(map(row))
        return cell

    return FieldMap_1_create_Z3BCCB7EB(_arrow282)


def FieldMap_1_field_298D8758(map: Callable[[_T], float]) -> FieldMap_1[Any]:
    def _arrow283(row: Any, cell: FsCell, map: Any=map) -> FsCell:
        cell.SetValueAs(map(row))
        return cell

    return FieldMap_1_create_Z3BCCB7EB(_arrow283)


def FieldMap_1_field_Z78848E24(map: Callable[[_T], int | None]) -> FieldMap_1[Any]:
    def _arrow284(row: Any, cell: FsCell, map: Any=map) -> FsCell:
        cell.SetValueAs(to_nullable(map(row)))
        return cell

    return FieldMap_1_create_Z3BCCB7EB(_arrow284)


def FieldMap_1_field_4C14CE8F(map: Callable[[_T], Any | None]) -> FieldMap_1[Any]:
    def _arrow285(row: Any, cell: FsCell, map: Any=map) -> FsCell:
        cell.SetValueAs(to_nullable(map(row)))
        return cell

    return FieldMap_1_create_Z3BCCB7EB(_arrow285)


def FieldMap_1_field_Z404517D6(map: Callable[[_T], bool | None]) -> FieldMap_1[Any]:
    def _arrow286(row: Any, cell: FsCell, map: Any=map) -> FsCell:
        cell.SetValueAs(to_nullable(map(row)))
        return cell

    return FieldMap_1_create_Z3BCCB7EB(_arrow286)


def FieldMap_1_field_Z14464045(map: Callable[[_T], float | None]) -> FieldMap_1[Any]:
    def _arrow287(row: Any, cell: FsCell, map: Any=map) -> FsCell:
        cell.SetValueAs(to_nullable(map(row)))
        return cell

    return FieldMap_1_create_Z3BCCB7EB(_arrow287)


def FieldMap_1_field_33F53BB(map: Callable[[_T], str | None]) -> FieldMap_1[Any]:
    def _arrow288(row: Any, cell: FsCell, map: Any=map) -> FsCell:
        match_value: str | None = map(row)
        if match_value is not None:
            text: str = match_value
            cell.SetValueAs(text)
            return cell

        else: 
            return cell


    return FieldMap_1_create_Z3BCCB7EB(_arrow288)


def FsSpreadsheet_FsTable__FsTable_Populate_526F9CF7(self_1: FsTable, cells: FsCellsCollection, data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> None:
    def mapping(field: FieldMap_1[_T], self_1: Any=self_1, cells: Any=cells, data: Any=data, fields: Any=fields) -> FSharpList[Callable[[_T, FsCell], FsCell]]:
        return field.HeaderTransformers

    headers_available: bool = not is_empty(concat(map_1(mapping, fields)))
    if (self_1.ShowHeaderRow == False) if headers_available else False:
        self_1.ShowHeaderRow = headers_available

    start_address: FsAddress = self_1.RangeAddress.FirstAddress
    start_row_index: int = ((start_address.RowNumber + 1) if headers_available else start_address.RowNumber) or 0
    with get_enumerator(indexed(data)) as enumerator:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            for_loop_var: tuple[int, _T] = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            row: _T = for_loop_var[1]
            active_row_index: int = (for_loop_var[0] + start_row_index) or 0
            with get_enumerator(fields) as enumerator_1:
                while enumerator_1.System_Collections_IEnumerator_MoveNext():
                    field_1: FieldMap_1[_T] = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current()
                    header_cell: FsCell = FsCell.create_empty()
                    with get_enumerator(field_1.HeaderTransformers) as enumerator_2:
                        while enumerator_2.System_Collections_IEnumerator_MoveNext():
                            ignore(enumerator_2.System_Collections_Generic_IEnumerator_1_get_Current()(row)(header_cell))
                    header_string: str = field_1.Hash if (header_cell.ValueAsString() == "") else header_cell.ValueAsString()
                    table_field: FsTableField = self_1.Field(header_string, cells)
                    active_cell: FsCell = table_field.Column.Cell(active_row_index, cells)
                    with get_enumerator(field_1.CellTransformers) as enumerator_3:
                        while enumerator_3.System_Collections_IEnumerator_MoveNext():
                            ignore(enumerator_3.System_Collections_Generic_IEnumerator_1_get_Current()(row)(active_cell))


def FsSpreadsheet_FsTable__FsTable_populate_Static_Z735E4C44(table: FsTable, cells: FsCellsCollection, data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> None:
    FsSpreadsheet_FsTable__FsTable_Populate_526F9CF7(table, cells, data, fields)


def FsSpreadsheet_FsWorksheet__FsWorksheet_Populate_Z2A1350BF(self_1: FsWorksheet, data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> None:
    def mapping(field: FieldMap_1[_T], self_1: Any=self_1, data: Any=data, fields: Any=fields) -> FSharpList[Callable[[_T, FsCell], FsCell]]:
        return field.HeaderTransformers

    headers_available: bool = not is_empty(concat(map_1(mapping, fields)))
    headers: Any = dict([])
    with get_enumerator(indexed(data)) as enumerator:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            for_loop_var: tuple[int, _T] = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            row: _T = for_loop_var[1]
            start_row_index: int = (2 if headers_available else 1) or 0
            active_row: FsRow = self_1.Row(for_loop_var[0] + start_row_index)
            with get_enumerator(fields) as enumerator_1:
                while enumerator_1.System_Collections_IEnumerator_MoveNext():
                    field_1: FieldMap_1[_T] = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current()
                    header_cell: FsCell = FsCell.create_empty()
                    with get_enumerator(field_1.HeaderTransformers) as enumerator_2:
                        while enumerator_2.System_Collections_IEnumerator_MoveNext():
                            ignore(enumerator_2.System_Collections_Generic_IEnumerator_1_get_Current()(row)(header_cell))
                    index: int
                    pattern_input: tuple[bool, str] = ((False, field_1.Hash)) if equals(header_cell.Value, "") else ((True, header_cell.ValueAsString()))
                    header_string: str = pattern_input[1]
                    match_value: int | None = Dictionary_tryGetValue(header_string, headers)
                    if match_value is None:
                        i: int = (len(headers) + 1) or 0
                        add_to_dict(headers, header_string, i)
                        if pattern_input[0]:
                            value: None = self_1.Row(1).Item(i).CopyFrom(header_cell)
                            ignore(None)

                        index = i

                    else: 
                        index = match_value

                    active_cell: FsCell = active_row.Item(index)
                    with get_enumerator(field_1.CellTransformers) as enumerator_3:
                        while enumerator_3.System_Collections_IEnumerator_MoveNext():
                            ignore(enumerator_3.System_Collections_Generic_IEnumerator_1_get_Current()(row)(active_cell))
    self_1.SortRows()


def FsSpreadsheet_FsWorksheet__FsWorksheet_populate_Static_Z2578A106(sheet: FsWorksheet, data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> None:
    FsSpreadsheet_FsWorksheet__FsWorksheet_Populate_Z2A1350BF(sheet, data, fields)


def FsSpreadsheet_FsWorksheet__FsWorksheet_createFrom_Static_Z2DEFA746(name: str, data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> FsWorksheet:
    sheet: FsWorksheet = FsWorksheet(name)
    FsSpreadsheet_FsWorksheet__FsWorksheet_populate_Static_Z2578A106(sheet, data, fields)
    return sheet


def FsSpreadsheet_FsWorksheet__FsWorksheet_createFrom_Static_Z2A1350BF(data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> FsWorksheet:
    return FsSpreadsheet_FsWorksheet__FsWorksheet_createFrom_Static_Z2DEFA746("Sheet1", data, fields)


def FsSpreadsheet_FsWorksheet__FsWorksheet_PopulateTable_59F260F9(self_1: FsWorksheet, table_name: str, start_address: FsAddress, data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> None:
    def mapping(field: FieldMap_1[_T], self_1: Any=self_1, table_name: Any=table_name, start_address: Any=start_address, data: Any=data, fields: Any=fields) -> FSharpList[Callable[[_T, FsCell], FsCell]]:
        return field.HeaderTransformers

    headers_available: bool = not is_empty(concat(map_1(mapping, fields)))
    FsSpreadsheet_FsTable__FsTable_Populate_526F9CF7(self_1.Table(table_name, FsRangeAddress(start_address, start_address), headers_available), self_1.CellCollection, data, fields)
    self_1.SortRows()


def FsSpreadsheet_FsWorksheet__FsWorksheet_createTableFrom_Static_30234BE1(name: str, table_name: str, data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> FsWorksheet:
    sheet: FsWorksheet = FsWorksheet(name)
    FsSpreadsheet_FsWorksheet__FsWorksheet_PopulateTable_59F260F9(sheet, table_name, FsAddress(1, 1), data, fields)
    return sheet


def FsSpreadsheet_FsWorkbook__FsWorkbook_Populate_Z2DEFA746(self_1: FsWorkbook, name: str, data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> None:
    ignore(self_1.InitWorksheet(name))
    def predicate(s: FsWorksheet, self_1: Any=self_1, name: Any=name, data: Any=data, fields: Any=fields) -> bool:
        return s.Name == name

    FsSpreadsheet_FsWorksheet__FsWorksheet_populate_Static_Z2578A106(find(predicate, self_1.GetWorksheets()), data, fields)


def FsSpreadsheet_FsWorkbook__FsWorkbook_populate_Static_Z7163EB39(workbook: FsWorkbook, name: str, data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> None:
    FsSpreadsheet_FsWorkbook__FsWorkbook_Populate_Z2DEFA746(workbook, name, data, fields)


def FsSpreadsheet_FsWorkbook__FsWorkbook_createFrom_Static_Z2DEFA746(name: str, data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> FsWorkbook:
    workbook: FsWorkbook = FsWorkbook()
    FsSpreadsheet_FsWorkbook__FsWorkbook_populate_Static_Z7163EB39(workbook, name, data, fields)
    return workbook


def FsSpreadsheet_FsWorkbook__FsWorkbook_createFrom_Static_Z2A1350BF(data: IEnumerable_1[Any], fields: FSharpList[FieldMap_1[Any]]) -> FsWorkbook:
    return FsSpreadsheet_FsWorkbook__FsWorkbook_createFrom_Static_Z2DEFA746("Sheet1", data, fields)


def FsSpreadsheet_FsWorkbook__FsWorkbook_createFrom_Static_Z2E36FAE7(sheets: FSharpList[FsWorksheet]) -> FsWorkbook:
    workbook: FsWorkbook = FsWorkbook()
    def action(sheet: FsWorksheet, sheets: Any=sheets) -> None:
        value: None = workbook.AddWorksheet(sheet)
        ignore(None)

    iterate(action, sheets)
    return workbook


__all__ = ["Dictionary_tryGetValue", "Dictionary_length", "FieldMap_1_reflection", "FieldMap_1_empty", "FieldMap_1_create_Z3BCCB7EB", "FieldMap_1__header_Z721C83C5", "FieldMap_1__header_54F19B58", "FieldMap_1__adjustToContents", "FieldMap_1_field_Z5C94BCA1", "FieldMap_1_field_54F19B58", "FieldMap_1_field_477D79EC", "FieldMap_1_field_Z1D55A0D7", "FieldMap_1_field_298D8758", "FieldMap_1_field_Z78848E24", "FieldMap_1_field_4C14CE8F", "FieldMap_1_field_Z404517D6", "FieldMap_1_field_Z14464045", "FieldMap_1_field_33F53BB", "FsSpreadsheet_FsTable__FsTable_Populate_526F9CF7", "FsSpreadsheet_FsTable__FsTable_populate_Static_Z735E4C44", "FsSpreadsheet_FsWorksheet__FsWorksheet_Populate_Z2A1350BF", "FsSpreadsheet_FsWorksheet__FsWorksheet_populate_Static_Z2578A106", "FsSpreadsheet_FsWorksheet__FsWorksheet_createFrom_Static_Z2DEFA746", "FsSpreadsheet_FsWorksheet__FsWorksheet_createFrom_Static_Z2A1350BF", "FsSpreadsheet_FsWorksheet__FsWorksheet_PopulateTable_59F260F9", "FsSpreadsheet_FsWorksheet__FsWorksheet_createTableFrom_Static_30234BE1", "FsSpreadsheet_FsWorkbook__FsWorkbook_Populate_Z2DEFA746", "FsSpreadsheet_FsWorkbook__FsWorkbook_populate_Static_Z7163EB39", "FsSpreadsheet_FsWorkbook__FsWorkbook_createFrom_Static_Z2DEFA746", "FsSpreadsheet_FsWorkbook__FsWorkbook_createFrom_Static_Z2A1350BF", "FsSpreadsheet_FsWorkbook__FsWorkbook_createFrom_Static_Z2E36FAE7"]




from typing import Any

from .types import FSharpRef


def compare(x: int, y: int) -> int:
    return -1 if x < y else 1 if x > y else 0


def equals(a: int, b: int) -> bool:
    return a == b


def abs(x: int) -> int:
    return -x if x < 0 else x


def sign(x: int) -> int:
    return -1 if x < 0 else 1 if x > 0 else 0


def max(x: int, y: int) -> int:
    return x if x > y else y


def min(x: int, y: int) -> int:
    return x if x < y else y


def op_unary_negation(value: int) -> int:
    # Note that we cannot negate the smallest negative number
    return -value if value != -0x8000000000000000 else -0x8000000000000000


# def op_unary_negation(a: int) -> int: return -a
def op_unary_plus(a: int) -> int:
    return +a


def op_logical_not(a: int) -> int:
    return ~a


def op_addition(a: int, b: int) -> int:
    return a + b


def op_subtraction(a: int, b: int) -> int:
    return a - b


def op_multiply(a: int, b: int) -> int:
    return a * b


def op_division(a: int, b: int) -> int:
    return a // b


def op_modulus(a: int, b: int) -> int:
    return a % b


def op_right_shift(a: int, b: int) -> int:
    return a >> b


def op_left_shift(a: int, b: int) -> int:
    return a << b


def op_bitwise_and(a: int, b: int) -> int:
    return a & b


def op_bitwise_or(a: int, b: int) -> int:
    return a | b


def op_exclusive_or(a: int, b: int) -> int:
    return a ^ b


def op_less_than(a: int, b: int) -> bool:
    return a < b


def op_less_than_or_equal(a: int, b: int) -> bool:
    return a <= b


def op_greater_than(a: int, b: int) -> bool:
    return a > b


def op_greater_than_or_equal(a: int, b: int) -> bool:
    return a >= b


def op_equality(a: int, b: int) -> bool:
    return a == b


def op_inequality(a: int, b: int) -> bool:
    return a != b


def from_bits(lowBits: int, highBits: int, unsigned: bool) -> int:
    ret = lowBits + (highBits << 32)
    if not unsigned and ret > 0x7FFFFFFFFFFFFFFF:
        return ret - 0x10000000000000000

    return ret


def from_int(value: int, unsigned: bool = False) -> int:
    if unsigned and value < 0:
        return value + 0x10000000000000000
    return value


def from_value(value: Any, unsigned: bool = False) -> int:
    value = int(value)
    if unsigned and value < 0:
        return value + 0x10000000000000000
    elif not unsigned and value > 9223372036854775807:
        return value - 0x10000000000000000

    return value


def from_number(value: int, unsigned: bool) -> int:
    return int(value)


def to_number(value: int) -> float:
    return float(value)


def from_integer(value: int, unsigned: bool | None = None, kind: int | None = None) -> int:
    if unsigned and value < 0:
        return value + 0x10000000000000000
    elif not unsigned and value > 9223372036854775807:
        return value - 0x10000000000000000
    return value


def get_range(unsigned: bool) -> tuple[int, int]:
    return (0, 18446744073709551615) if unsigned else (-9223372036854775808, 9223372036854775807)


AllowHexSpecifier = 0x00000200


def parse(string: str, style: int, unsigned: bool, bitsize: int, radix: int = 10) -> int:
    # const res = isValid(str, style, radix);
    if style & AllowHexSpecifier or string.startswith("0x"):
        radix = 16
    elif string.startswith("0b"):
        radix = 2
    elif string.startswith("0o"):
        radix = 8

    try:
        v = int(string, base=radix)
    except Exception:
        raise ValueError(f"The input string {string} was not in a correct format.")

    (umin, umax) = get_range(True)
    if not unsigned and radix != 10 and v >= umin and v <= umax:
        mask = 1 << (bitsize - 1)
        if v & mask:  # Test if negative
            v = v - (mask << 1)

    (min, max) = get_range(unsigned)
    if v >= min and v <= max:
        return v

    raise ValueError(f"The input string {string} was not in a correct format.")


def try_parse(string: str, style: int, unsigned: bool, bitsize: int, defValue: FSharpRef[int]) -> bool:
    try:
        defValue.contents = parse(string, style, unsigned, bitsize)
        return True
    except Exception:
        return False


def to_string(x: int) -> str:
    return str(x)


def to_int(value: int) -> int:
    if value > 9223372036854775807:
        return value - 0x10000000000000000
    return value


long = int

__all__ = [
    "compare",
    "equals",
    "abs",
    "sign",
    "max",
    "min",
    "op_unary_negation",
    "op_unary_plus",
    "op_logical_not",
    "op_addition",
    "op_subtraction",
    "op_multiply",
    "op_division",
    "op_modulus",
    "op_right_shift",
    "op_left_shift",
    "op_bitwise_and",
    "op_bitwise_or",
    "op_exclusive_or",
    "op_less_than",
    "op_less_than_or_equal",
    "op_greater_than",
    "op_greater_than_or_equal",
    "op_equality",
    "op_inequality",
    "from_bits",
    "from_int",
    "from_value",
    "from_number",
    "to_number",
    "from_integer",
    "get_range",
    "parse",
    "try_parse",
    "to_string",
    "to_int",
    "long",
]



from __future__ import annotations
from typing import Any
from ...arctrl_core.data_context import (DataContext__ctor_Z780A8A2A, DataContext)
from ...arctrl_core.data_map import DataMap
from ...fable_library.list import (of_array, FSharpList, exists, iterate, length, head, iterate_indexed)
from ...fable_library.map_util import add_to_dict
from ...fable_library.range import range_big_int
from ...fable_library.seq import (try_find, item, to_array, delay, map, to_list)
from ...fable_library.string_ import (starts_with_exact, to_fail, printf)
from ...fable_library.types import Array
from ...fable_library.util import (IEnumerable_1, ignore)
from ...fs_spreadsheet.Cells.fs_cell import FsCell
from ...fs_spreadsheet.Cells.fs_cells_collection import Dictionary_tryGet
from ...fs_spreadsheet.fs_address import FsAddress
from ...fs_spreadsheet.fs_column import FsColumn
from ...fs_spreadsheet.fs_worksheet import FsWorksheet
from ...fs_spreadsheet.Ranges.fs_range_address import FsRangeAddress
from ...fs_spreadsheet.Tables.fs_table import FsTable
from ..AnnotationTable.arc_table import Aux_List_groupWhen
from .data_map_column import (set_from_fs_columns, to_fs_columns)

helper_column_strings: FSharpList[str] = of_array(["Term Source REF", "Term Accession Number", "Data Format", "Data Selector Format"])

def group_columns_by_header(columns: FSharpList[FsColumn]) -> FSharpList[FSharpList[FsColumn]]:
    def f(c: FsColumn, columns: Any=columns) -> bool:
        v: str = c.Item(1).ValueAsString()
        def predicate(s: str, c: Any=c) -> bool:
            return starts_with_exact(v, s)

        return not exists(predicate, helper_column_strings)

    return Aux_List_groupWhen(f, columns)


def try_data_map_table(sheet: FsWorksheet) -> FsTable | None:
    def predicate(t: FsTable, sheet: Any=sheet) -> bool:
        return starts_with_exact(t.Name, "datamapTable")

    return try_find(predicate, sheet.Tables)


def compose_columns(columns: IEnumerable_1[FsColumn]) -> Array[DataContext]:
    l: int = (item(0, columns).MaxRowIndex - 1) or 0
    def _arrow3633(__unit: None=None, columns: Any=columns) -> IEnumerable_1[DataContext]:
        def _arrow3632(i: int) -> DataContext:
            return DataContext__ctor_Z780A8A2A()

        return map(_arrow3632, range_big_int(0, 1, l - 1))

    dc: Array[DataContext] = list(to_array(delay(_arrow3633)))
    def action(arg: FSharpList[FsColumn], columns: Any=columns) -> None:
        ignore(set_from_fs_columns(dc, arg))

    iterate(action, group_columns_by_header(to_list(columns)))
    return dc


def try_from_fs_worksheet(sheet: FsWorksheet) -> DataMap | None:
    try: 
        match_value: FsTable | None = try_data_map_table(sheet)
        if match_value is None:
            return None

        else: 
            t: FsTable = match_value
            return DataMap(compose_columns(t.GetColumns(sheet.CellCollection)))


    except Exception as err:
        arg: str = sheet.Name
        arg_1: str = str(err)
        return to_fail(printf("Could not parse datamap table with name \"%s\":\n%s"))(arg)(arg_1)



def to_fs_worksheet(table: DataMap) -> FsWorksheet:
    string_count: Any = dict([])
    ws: FsWorksheet = FsWorksheet("isa_datamap")
    if len(table.DataContexts) == 0:
        return ws

    else: 
        columns: FSharpList[FSharpList[FsCell]] = to_fs_columns(table.DataContexts)
        max_row: int = length(head(columns)) or 0
        max_col: int = length(columns) or 0
        fs_table: FsTable = ws.Table("datamapTable", FsRangeAddress(FsAddress(1, 1), FsAddress(max_row, max_col)))
        def action_1(col_i: int, col: FSharpList[FsCell], table: Any=table) -> None:
            def action(row_i: int, cell: FsCell, col_i: Any=col_i, col: Any=col) -> None:
                value: str
                v: str = cell.ValueAsString()
                if row_i == 0:
                    match_value: str | None = Dictionary_tryGet(v, string_count)
                    if match_value is None:
                        add_to_dict(string_count, cell.ValueAsString(), "")
                        value = v

                    else: 
                        spaces: str = match_value
                        string_count[v] = spaces + " "
                        value = (v + " ") + spaces


                else: 
                    value = v

                address: FsAddress = FsAddress(row_i + 1, col_i + 1)
                fs_table.Cell(address, ws.CellCollection).SetValueAs(value)

            iterate_indexed(action, col)

        iterate_indexed(action_1, columns)
        ws.RescanRows()
        return ws



__all__ = ["helper_column_strings", "group_columns_by_header", "try_data_map_table", "compose_columns", "try_from_fs_worksheet", "to_fs_worksheet"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..fable_library.list import (FSharpList, is_empty, tail, head, of_seq as of_seq_1, empty, cons, reverse)
from ..fable_library.option import value as value_1
from ..fable_library.range import range_big_int
from ..fable_library.seq import (to_list, delay, map)
from ..fable_library.string_ import (join, to_console, printf)
from ..fable_library.types import Array
from ..fable_library.util import (IEnumerable_1, get_enumerator, ignore)
from .dynamic_obj import (DynamicObj, CopyUtils_tryDeepCopyObj_75B3D832)
from .property_helper import PropertyHelper

_TPROPERTYVALUE = TypeVar("_TPROPERTYVALUE")

_UPROPERTYVALUE = TypeVar("_UPROPERTYVALUE")

def of_dict(dynamic_properties: Any) -> DynamicObj:
    return DynamicObj.of_dict(dynamic_properties)


def of_seq(dynamic_properties: IEnumerable_1[tuple[str, Any]]) -> DynamicObj:
    dynamic_properties_1: Any = dict(dict(dynamic_properties))
    return DynamicObj.of_dict(dynamic_properties_1)


def of_list(dynamic_properties: FSharpList[tuple[str, Any]]) -> DynamicObj:
    return of_seq(dynamic_properties)


def of_array(dynamic_properties: Array[tuple[str, Any]]) -> DynamicObj:
    return of_seq(dynamic_properties)


def combine(first: DynamicObj, second: DynamicObj) -> DynamicObj:
    with get_enumerator(second.GetProperties(True)) as enumerator:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            kv: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            match_value: Any = kv[1]
            if isinstance(match_value, DynamicObj):
                match_value_1: Any | None = first.TryGetPropertyValue(kv[0])
                if match_value_1 is None:
                    first.SetProperty(kv[0], match_value)

                else: 
                    tmp: DynamicObj = combine(value_1(match_value_1), match_value)
                    first.SetProperty(kv[0], tmp)


            else: 
                first.SetProperty(kv[0], kv[1])

    return first


def set_property(property_name: str, property_value: Any, dyn_obj: DynamicObj) -> None:
    dyn_obj.SetProperty(property_name, property_value)


def set_optional_property(property_name: str, property_value: Any | None, dyn_obj: DynamicObj) -> None:
    if property_value is None:
        pass

    else: 
        set_property(property_name, value_1(property_value), dyn_obj)



def set_optional_property_by(property_name: str, property_value: Any | None, mapping: Callable[[_TPROPERTYVALUE], _UPROPERTYVALUE], dyn_obj: DynamicObj) -> None:
    if property_value is None:
        pass

    else: 
        set_property(property_name, mapping(value_1(property_value)), dyn_obj)



def try_get_property_value(property_name: str, dyn_obj: DynamicObj) -> Any | None:
    return dyn_obj.TryGetPropertyValue(property_name)


def remove_property(property_name: str, dyn_obj: DynamicObj) -> None:
    ignore(dyn_obj.RemoveProperty(property_name))


def format(dyn_obj: DynamicObj) -> str:
    def loop(object_mut: DynamicObj, indentation_level_mut: int, members_left_mut: FSharpList[PropertyHelper], acc_mut: FSharpList[str], dyn_obj: Any=dyn_obj) -> str:
        while True:
            (object, indentation_level, members_left, acc) = (object_mut, indentation_level_mut, members_left_mut, acc_mut)
            def _arrow56(__unit: None=None, object: Any=object, indentation_level: Any=indentation_level, members_left: Any=members_left, acc: Any=acc) -> IEnumerable_1[str]:
                def _arrow55(i: int) -> str:
                    return "    "

                return map(_arrow55, range_big_int(0, 1, indentation_level - 1))

            indent: str = join("", to_list(delay(_arrow56)))
            if not is_empty(members_left):
                rest: FSharpList[PropertyHelper] = tail(members_left)
                m: PropertyHelper = head(members_left)
                item: Any = m.GetValue(object)
                dynamic_indicator: str = "?" if m.IsDynamic else ""
                name: str = m.Name
                if isinstance(item, DynamicObj):
                    object_mut = object
                    indentation_level_mut = indentation_level
                    members_left_mut = rest
                    acc_mut = cons(((((((((("" + indent) + "") + dynamic_indicator) + "") + name) + ":") + "\n") + "") + loop(item, indentation_level + 1, of_seq_1(item.GetPropertyHelpers(True)), empty())) + "", acc)
                    continue

                else: 
                    object_mut = object
                    indentation_level_mut = indentation_level
                    members_left_mut = rest
                    acc_mut = cons(((((((("" + indent) + "") + dynamic_indicator) + "") + name) + ": ") + str(item)) + "", acc)
                    continue


            else: 
                return join("\n", reverse(acc))

            break

    return loop(dyn_obj, 0, of_seq_1(dyn_obj.GetPropertyHelpers(True)), empty())


def print(dyn_obj: DynamicObj) -> None:
    arg: str = format(dyn_obj)
    to_console(printf("%s"))(arg)


def try_deep_copy_obj(include_instance_properties: bool, o: DynamicObj) -> Any:
    return CopyUtils_tryDeepCopyObj_75B3D832(o, include_instance_properties)


__all__ = ["of_dict", "of_seq", "of_list", "of_array", "combine", "set_property", "set_optional_property", "set_optional_property_by", "try_get_property_value", "remove_property", "format", "print", "try_deep_copy_obj"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..fable_library.date import to_string as to_string_1
from ..fable_library.list import (of_seq, map as map_1, of_array, FSharpList)
from ..fable_library.option import value as value_1
from ..fable_library.seq import map
from ..fable_library.types import (to_string, Array)
from ..fable_library.util import (int32_to_string, IEnumerable_1)
from .writer import write as write_1
from .yamlicious_types import (YAMLContent_create_27AED5E3, YAMLElement, Config)

_A = TypeVar("_A")

def int_1(value: int) -> YAMLElement:
    return YAMLElement(1, YAMLContent_create_27AED5E3(int32_to_string(value)))


def float_1(value: float) -> YAMLElement:
    return YAMLElement(1, YAMLContent_create_27AED5E3(str(value)))


def char(value: str) -> YAMLElement:
    return YAMLElement(1, YAMLContent_create_27AED5E3(value))


def bool_1(value: bool) -> YAMLElement:
    return YAMLElement(1, YAMLContent_create_27AED5E3(to_string(value)))


def datetime(value: Any) -> YAMLElement:
    return YAMLElement(1, YAMLContent_create_27AED5E3(to_string_1(value, "O", {})))


def datetime_offset(value: Any) -> YAMLElement:
    return YAMLElement(1, YAMLContent_create_27AED5E3(to_string_1(value, "O", {})))


def string(value: str) -> YAMLElement:
    return YAMLElement(1, YAMLContent_create_27AED5E3(value))


def seq(encoder: Callable[[_A], YAMLElement], s: IEnumerable_1[Any]) -> YAMLElement:
    return YAMLElement(2, of_seq(map(encoder, s)))


def array(encoder: Callable[[_A], YAMLElement], arr: Array[Any]) -> YAMLElement:
    return YAMLElement(2, map_1(encoder, of_array(arr)))


def resizearray(encoder: Callable[[_A], YAMLElement], arr: Array[Any]) -> YAMLElement:
    return YAMLElement(2, map_1(encoder, of_seq(arr)))


def list_1(encoder: Callable[[_A], YAMLElement], l: FSharpList[Any]) -> YAMLElement:
    return YAMLElement(2, map_1(encoder, l))


def values(encoder: Callable[[_A], str], values_1: IEnumerable_1[Any]) -> YAMLElement:
    def mapping(arg_1: _A | None=None, encoder: Any=encoder, values_1: Any=values_1) -> YAMLElement:
        return YAMLElement(1, YAMLContent_create_27AED5E3(encoder(arg_1)))

    return YAMLElement(3, of_seq(map(mapping, values_1)))


def comment(string_1: str) -> YAMLElement:
    return YAMLElement(4, string_1)


nil: YAMLElement = YAMLElement(5)

def try_include(name: str, encoder: Callable[[_A], YAMLElement], value: Any | None=None) -> tuple[str, YAMLElement]:
    return (name, encoder(value_1(value)) if (value is not None) else nil)


def write(whitespaces: int, ele: YAMLElement) -> str:
    def _arrow411(c: Config, whitespaces: Any=whitespaces, ele: Any=ele) -> Config:
        return Config(whitespaces, c.Level)

    return write_1(ele, _arrow411)


__all__ = ["int_1", "float_1", "char", "bool_1", "datetime", "datetime_offset", "string", "seq", "array", "resizearray", "list_1", "values", "comment", "nil", "try_include", "write"]




from __future__ import annotations
from datetime import datetime
from typing import Any
from ...fable_library.option import value
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)
from .comment import LDComment
from .file import LDFile
from .lab_process import LDLabProcess
from .person import LDPerson
from .scholarly_article import LDScholarlyArticle

def _expr1531() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDCreativeWork", None, LDCreativeWork)


class LDCreativeWork:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/CreativeWork"

    @staticmethod
    def identifier() -> str:
        return "http://schema.org/identifier"

    @staticmethod
    def creator() -> str:
        return "http://schema.org/creator"

    @staticmethod
    def date_created() -> str:
        return "http://schema.org/dateCreated"

    @staticmethod
    def date_published() -> str:
        return "http://schema.org/datePublished"

    @staticmethod
    def sd_date_published() -> str:
        return "http://schema.org/datePublished"

    @staticmethod
    def license() -> str:
        return "http://schema.org/license"

    @staticmethod
    def date_modified() -> str:
        return "http://schema.org/dateModified"

    @staticmethod
    def description() -> str:
        return "http://schema.org/description"

    @staticmethod
    def has_part() -> str:
        return "http://schema.org/hasPart"

    @staticmethod
    def headline() -> str:
        return "http://schema.org/headline"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def text() -> str:
        return "http://schema.org/text"

    @staticmethod
    def citation() -> str:
        return "http://schema.org/citation"

    @staticmethod
    def comment() -> str:
        return "http://schema.org/comment"

    @staticmethod
    def mentions() -> str:
        return "http://schema.org/mentions"

    @staticmethod
    def url() -> str:
        return "http://schema.org/url"

    @staticmethod
    def about() -> str:
        return "http://schema.org/about"

    @staticmethod
    def try_get_identifier_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.identifier(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_identifier_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.identifier(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("property `identifier` of object with @id `" + lp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `identifier` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_identifier_as_string(lp: LDNode, identifier: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.identifier(), identifier, context)

    @staticmethod
    def get_creators(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(p: LDNode, ctx: LDContext | None=None) -> bool:
            return LDPerson.validate(p, ctx)

        return lp.GetPropertyNodes(LDCreativeWork.creator(), filter, graph, context)

    @staticmethod
    def set_creators(lp: LDNode, creators: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.creator(), creators, context)

    @staticmethod
    def try_get_date_created_as_date_time(lp: LDNode, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.date_created(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if isinstance(value(match_value), datetime):
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_date_created_as_date_time(lp: LDNode, date_created: Any, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.date_created(), date_created, context)

    @staticmethod
    def try_get_date_published_as_date_time(lp: LDNode, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.date_published(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if isinstance(value(match_value), datetime):
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_date_published_as_date_time(lp: LDNode, date_published: Any, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.date_published(), date_published, context)

    @staticmethod
    def try_get_sddate_published_as_date_time(lp: LDNode, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.sd_date_published(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if isinstance(value(match_value), datetime):
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_sddate_published_as_date_time(lp: LDNode, sd_date_published: Any, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.sd_date_published(), sd_date_published, context)

    @staticmethod
    def try_get_license_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.license(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_license_as_string(lp: LDNode, license: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.license(), license, context)

    @staticmethod
    def set_license_as_creative_work(lp: LDNode, license: Any=None, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.license(), license, context)

    @staticmethod
    def try_get_date_modified_as_date_time(lp: LDNode, context: LDContext | None=None) -> Any | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.date_modified(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if isinstance(value(match_value), datetime):
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_date_modified_as_date_time(lp: LDNode, date_modified: Any, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.date_modified(), date_modified, context)

    @staticmethod
    def try_get_description_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.description(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_description_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.description(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("property `description` of object with @id `" + lp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `description` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_description_as_string(lp: LDNode, description: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.description(), description, context)

    @staticmethod
    def get_has_parts(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return lp.GetPropertyNodes(LDCreativeWork.has_part(), None, graph, context)

    @staticmethod
    def get_has_parts_as_dataset(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDCreativeWork.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDCreativeWork.has_part(), filter, graph, context)

    @staticmethod
    def get_has_parts_as_file(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDFile.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDCreativeWork.has_part(), filter, graph, context)

    @staticmethod
    def set_has_parts(lp: LDNode, has_parts: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.has_part(), has_parts, context)

    @staticmethod
    def try_get_headline_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.headline(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def try_get_name_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.name(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("property `name` of object with @id `" + lp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `name` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_name_as_string(lp: LDNode, name: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.name(), name, context)

    @staticmethod
    def try_get_text_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.text(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_text_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.text(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("property `text` of object with @id `" + lp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `text` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_text_as_string(lp: LDNode, text: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.text(), text, context)

    @staticmethod
    def get_citations(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDScholarlyArticle.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDCreativeWork.citation(), filter, graph, context)

    @staticmethod
    def set_citations(lp: LDNode, citations: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.citation(), citations, context)

    @staticmethod
    def get_comments(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDComment.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDCreativeWork.comment(), filter, graph, context)

    @staticmethod
    def set_comments(lp: LDNode, comments: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.comment(), comments, context)

    @staticmethod
    def try_get_url_as_string(lp: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.url(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_url_as_string(lp: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = lp.TryGetPropertyAsSingleton(LDCreativeWork.url(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("property `url` of object with @id `" + lp.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `url` of object with @id `" + lp.Id) + "`")


    @staticmethod
    def set_url_as_string(lp: LDNode, url: str, context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.url(), url, context)

    @staticmethod
    def get_abouts(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return lp.GetPropertyNodes(LDCreativeWork.about(), None, graph, context)

    @staticmethod
    def get_abouts_as_lab_process(lp: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ldnode: LDNode, context_1: LDContext | None=None) -> bool:
            return LDLabProcess.validate(ldnode, context_1)

        return lp.GetPropertyNodes(LDCreativeWork.about(), filter, graph, context)

    @staticmethod
    def set_abouts(lp: LDNode, abouts: Array[LDNode], context: LDContext | None=None) -> Any:
        return lp.SetProperty(LDCreativeWork.about(), abouts, context)

    @staticmethod
    def validate(lp: LDNode, context: LDContext | None=None) -> bool:
        return lp.HasType(LDCreativeWork.schema_type(), context)

    @staticmethod
    def create(id: str, identifier: str | None=None, creators: Array[LDNode] | None=None, date_created: Any | None=None, date_published: Any | None=None, date_modified: Any | None=None, description: str | None=None, has_parts: Array[LDNode] | None=None, name: str | None=None, text: str | None=None, citations: Array[LDNode] | None=None, comments: Array[LDNode] | None=None, mentions: Array[LDNode] | None=None, url: str | None=None, abouts: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        s: LDNode = LDNode(id, [LDCreativeWork.schema_type()], None, context)
        s.SetOptionalProperty(LDCreativeWork.identifier(), identifier, context)
        s.SetOptionalProperty(LDCreativeWork.creator(), creators, context)
        s.SetOptionalProperty(LDCreativeWork.date_created(), date_created, context)
        s.SetOptionalProperty(LDCreativeWork.date_published(), date_published, context)
        s.SetOptionalProperty(LDCreativeWork.date_modified(), date_modified, context)
        s.SetOptionalProperty(LDCreativeWork.description(), description, context)
        s.SetOptionalProperty(LDCreativeWork.has_part(), has_parts, context)
        s.SetOptionalProperty(LDCreativeWork.name(), name, context)
        s.SetOptionalProperty(LDCreativeWork.text(), text, context)
        s.SetOptionalProperty(LDCreativeWork.citation(), citations, context)
        s.SetOptionalProperty(LDCreativeWork.comment(), comments, context)
        s.SetOptionalProperty(LDCreativeWork.mentions(), mentions, context)
        s.SetOptionalProperty(LDCreativeWork.url(), url, context)
        s.SetOptionalProperty(LDCreativeWork.about(), abouts, context)
        return s


LDCreativeWork_reflection = _expr1531

__all__ = ["LDCreativeWork_reflection"]




from __future__ import annotations
from typing import Any
from ...fable_library.option import default_arg
from ...fable_library.reflection import (TypeInfo, string_type, union_type)
from ...fable_library.types import (Array, Union)
from ..data import (Data, Data_reflection)
from ..ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)
from .composite_header import CompositeHeader

def _expr819() -> TypeInfo:
    return union_type("ARCtrl.CompositeCell", [], CompositeCell, lambda: [[("Item", OntologyAnnotation_reflection())], [("Item", string_type)], [("Item1", string_type), ("Item2", OntologyAnnotation_reflection())], [("Item", Data_reflection())]])


class CompositeCell(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Term", "FreeText", "Unitized", "Data"]

    @property
    def is_unitized(self, __unit: None=None) -> bool:
        this: CompositeCell = self
        return True if (this.tag == 2) else False

    @property
    def is_term(self, __unit: None=None) -> bool:
        this: CompositeCell = self
        return True if (this.tag == 0) else False

    @property
    def is_free_text(self, __unit: None=None) -> bool:
        this: CompositeCell = self
        return True if (this.tag == 1) else False

    @property
    def is_data(self, __unit: None=None) -> bool:
        this: CompositeCell = self
        return True if (this.tag == 3) else False

    def GetEmptyCell(self, __unit: None=None) -> CompositeCell:
        this: CompositeCell = self
        if this.tag == 2:
            return CompositeCell.empty_unitized()

        elif this.tag == 1:
            return CompositeCell.empty_free_text()

        elif this.tag == 3:
            return CompositeCell.empty_data()

        else: 
            return CompositeCell.empty_term()


    def GetContent(self, __unit: None=None) -> Array[str]:
        this: CompositeCell = self
        if this.tag == 0:
            oa: OntologyAnnotation = this.fields[0]
            return [oa.NameText, default_arg(oa.TermSourceREF, ""), default_arg(oa.TermAccessionNumber, "")]

        elif this.tag == 2:
            oa_1: OntologyAnnotation = this.fields[1]
            return [this.fields[0], oa_1.NameText, default_arg(oa_1.TermSourceREF, ""), default_arg(oa_1.TermAccessionNumber, "")]

        elif this.tag == 3:
            d: Data = this.fields[0]
            return [default_arg(d.Name, ""), default_arg(d.Format, ""), default_arg(d.SelectorFormat, "")]

        else: 
            return [this.fields[0]]


    def ToUnitizedCell(self, __unit: None=None) -> CompositeCell:
        this: CompositeCell = self
        if this.tag == 1:
            return CompositeCell(2, "", OntologyAnnotation.create(this.fields[0]))

        elif this.tag == 0:
            return CompositeCell(2, "", this.fields[0])

        elif this.tag == 3:
            return CompositeCell(2, "", OntologyAnnotation.create(this.fields[0].NameText))

        else: 
            return this


    def ToTermCell(self, __unit: None=None) -> CompositeCell:
        this: CompositeCell = self
        if this.tag == 2:
            return CompositeCell(0, this.fields[1])

        elif this.tag == 1:
            return CompositeCell(0, OntologyAnnotation.create(this.fields[0]))

        elif this.tag == 3:
            return CompositeCell(0, OntologyAnnotation(this.fields[0].NameText))

        else: 
            return this


    def ToFreeTextCell(self, __unit: None=None) -> CompositeCell:
        this: CompositeCell = self
        if this.tag == 0:
            return CompositeCell(1, this.fields[0].NameText)

        elif this.tag == 2:
            return CompositeCell(1, this.fields[1].NameText)

        elif this.tag == 3:
            return CompositeCell(1, this.fields[0].NameText)

        else: 
            return this


    def ToDataCell(self, __unit: None=None) -> CompositeCell:
        this: CompositeCell = self
        if this.tag == 1:
            return CompositeCell.create_data_from_string(this.fields[0])

        elif this.tag == 0:
            return CompositeCell.create_data_from_string(this.fields[0].NameText)

        elif this.tag == 3:
            return this

        else: 
            return CompositeCell.create_data_from_string(this.fields[1].NameText)


    @property
    def AsUnitized(self, __unit: None=None) -> tuple[str, OntologyAnnotation]:
        this: CompositeCell = self
        if this.tag == 2:
            return (this.fields[0], this.fields[1])

        else: 
            raise Exception("Not a Unitized cell.")


    @property
    def AsTerm(self, __unit: None=None) -> OntologyAnnotation:
        this: CompositeCell = self
        if this.tag == 0:
            return this.fields[0]

        else: 
            raise Exception("Not a Term Cell.")


    @property
    def AsFreeText(self, __unit: None=None) -> str:
        this: CompositeCell = self
        if this.tag == 1:
            return this.fields[0]

        else: 
            raise Exception("Not a FreeText Cell.")


    @property
    def AsData(self, __unit: None=None) -> Data:
        this: CompositeCell = self
        if this.tag == 3:
            return this.fields[0]

        else: 
            raise Exception("Not a Data Cell.")


    @staticmethod
    def create_term(oa: OntologyAnnotation) -> CompositeCell:
        return CompositeCell(0, oa)

    @staticmethod
    def create_term_from_string(name: str | None=None, tsr: str | None=None, tan: str | None=None) -> CompositeCell:
        return CompositeCell(0, OntologyAnnotation.create(name, tsr, tan))

    @staticmethod
    def create_unitized(value: str, oa: OntologyAnnotation | None=None) -> CompositeCell:
        return CompositeCell(2, value, default_arg(oa, OntologyAnnotation()))

    @staticmethod
    def create_unitized_from_string(value: str, name: str | None=None, tsr: str | None=None, tan: str | None=None) -> CompositeCell:
        tupled_arg: tuple[str, OntologyAnnotation] = (value, OntologyAnnotation.create(name, tsr, tan))
        return CompositeCell(2, tupled_arg[0], tupled_arg[1])

    @staticmethod
    def create_free_text(value: str) -> CompositeCell:
        return CompositeCell(1, value)

    @staticmethod
    def create_data(d: Data) -> CompositeCell:
        return CompositeCell(3, d)

    @staticmethod
    def create_data_from_string(value: str, format: str | None=None, selector_format: str | None=None) -> CompositeCell:
        return CompositeCell(3, Data.create(None, value, None, format, selector_format))

    @staticmethod
    def empty_term() -> CompositeCell:
        return CompositeCell(0, OntologyAnnotation())

    @staticmethod
    def empty_free_text() -> CompositeCell:
        return CompositeCell(1, "")

    @staticmethod
    def empty_unitized() -> CompositeCell:
        return CompositeCell(2, "", OntologyAnnotation())

    @staticmethod
    def empty_data() -> CompositeCell:
        return CompositeCell(3, Data.create())

    def UpdateWithOA(self, oa: OntologyAnnotation) -> CompositeCell:
        this: CompositeCell = self
        if this.tag == 2:
            return CompositeCell.create_unitized(this.fields[0], oa)

        elif this.tag == 1:
            return CompositeCell.create_free_text(oa.NameText)

        elif this.tag == 3:
            d: Data = this.fields[0]
            d.Name = oa.NameText
            return CompositeCell(3, d)

        else: 
            return CompositeCell.create_term(oa)


    @staticmethod
    def update_with_oa(oa: OntologyAnnotation, cell: CompositeCell) -> CompositeCell:
        return cell.UpdateWithOA(oa)

    def __str__(self, __unit: None=None) -> str:
        this: CompositeCell = self
        if this.tag == 1:
            return this.fields[0]

        elif this.tag == 2:
            return ((("" + this.fields[0]) + " ") + this.fields[1].NameText) + ""

        elif this.tag == 3:
            return ("" + this.fields[0].NameText) + ""

        else: 
            return ("" + this.fields[0].NameText) + ""


    def Copy(self, __unit: None=None) -> CompositeCell:
        this: CompositeCell = self
        if this.tag == 1:
            return CompositeCell(1, this.fields[0])

        elif this.tag == 2:
            return CompositeCell(2, this.fields[0], this.fields[1].Copy())

        elif this.tag == 3:
            return CompositeCell(3, this.fields[0].Copy())

        else: 
            return CompositeCell(0, this.fields[0].Copy())


    def ValidateAgainstHeader(self, header: CompositeHeader, raise_exception: bool | None=None) -> bool:
        this: CompositeCell = self
        raise_exeption: bool = default_arg(raise_exception, False)
        cell: CompositeCell = this
        if (True if cell.is_data else cell.is_free_text) if header.IsDataColumn else False:
            return True

        elif header.IsDataColumn:
            if raise_exeption:
                raise Exception(((("Invalid combination of header `" + str(header)) + "` and cell `") + str(cell)) + "`, Data header should have either Data or Freetext cells")

            return False

        elif (True if cell.is_term else cell.is_unitized) if header.IsTermColumn else False:
            return True

        elif cell.is_free_text if (not header.IsTermColumn) else False:
            return True

        else: 
            if raise_exeption:
                raise Exception(((("Invalid combination of header `" + str(header)) + "` and cell `") + str(cell)) + "`")

            return False


    @staticmethod
    def term(oa: OntologyAnnotation) -> CompositeCell:
        return CompositeCell(0, oa)

    @staticmethod
    def free_text(s: str) -> CompositeCell:
        return CompositeCell(1, s)

    @staticmethod
    def unitized(v: str, oa: OntologyAnnotation) -> CompositeCell:
        return CompositeCell(2, v, oa)

    @staticmethod
    def data(d: Data) -> CompositeCell:
        return CompositeCell(3, d)


CompositeCell_reflection = _expr819

__all__ = ["CompositeCell_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_json.encode import default_spaces
from ...arctrl_json.ontology_annotation import (OntologyAnnotation_decoder, OntologyAnnotation_encoder, OntologyAnnotation_ROCrate_decoderDefinedTerm, OntologyAnnotation_ROCrate_encoderDefinedTerm, OntologyAnnotation_ISAJson_decoder, OntologyAnnotation_ISAJson_encoder)
from ...fable_library.result import FSharpResult_2
from ...fable_library.string_ import (to_text, printf)
from ...thoth_json_core.types import IEncodable
from ...thoth_json_python.decode import Decode_fromString
from ...thoth_json_python.encode import to_string

def ARCtrl_OntologyAnnotation__OntologyAnnotation_fromJsonString_Static_Z721C83C5(s: str) -> OntologyAnnotation:
    match_value: FSharpResult_2[OntologyAnnotation, str] = Decode_fromString(OntologyAnnotation_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_OntologyAnnotation__OntologyAnnotation_toJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[OntologyAnnotation], str]:
    def _arrow3786(obj: OntologyAnnotation, spaces: Any=spaces) -> str:
        value: IEncodable = OntologyAnnotation_encoder(obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3786


def ARCtrl_OntologyAnnotation__OntologyAnnotation_ToJsonString_71136F3F(this: OntologyAnnotation, spaces: int | None=None) -> str:
    return ARCtrl_OntologyAnnotation__OntologyAnnotation_toJsonString_Static_71136F3F(spaces)(this)


def ARCtrl_OntologyAnnotation__OntologyAnnotation_fromROCrateJsonString_Static_Z721C83C5(s: str) -> OntologyAnnotation:
    match_value: FSharpResult_2[OntologyAnnotation, str] = Decode_fromString(OntologyAnnotation_ROCrate_decoderDefinedTerm, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_OntologyAnnotation__OntologyAnnotation_toROCrateJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[OntologyAnnotation], str]:
    def _arrow3787(obj: OntologyAnnotation, spaces: Any=spaces) -> str:
        value: IEncodable = OntologyAnnotation_ROCrate_encoderDefinedTerm(obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3787


def ARCtrl_OntologyAnnotation__OntologyAnnotation_ToROCrateJsonString_71136F3F(this: OntologyAnnotation, spaces: int | None=None) -> str:
    return ARCtrl_OntologyAnnotation__OntologyAnnotation_toROCrateJsonString_Static_71136F3F(spaces)(this)


def ARCtrl_OntologyAnnotation__OntologyAnnotation_fromISAJsonString_Static_Z721C83C5(s: str) -> OntologyAnnotation:
    match_value: FSharpResult_2[OntologyAnnotation, str] = Decode_fromString(OntologyAnnotation_ISAJson_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_OntologyAnnotation__OntologyAnnotation_toISAJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[OntologyAnnotation], str]:
    def _arrow3788(obj: OntologyAnnotation, spaces: Any=spaces) -> str:
        value: IEncodable = OntologyAnnotation_ISAJson_encoder(None, obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3788


def ARCtrl_OntologyAnnotation__OntologyAnnotation_ToISAJsonString_71136F3F(this: OntologyAnnotation, spaces: int | None=None) -> str:
    return ARCtrl_OntologyAnnotation__OntologyAnnotation_toISAJsonString_Static_71136F3F(spaces)(this)


__all__ = ["ARCtrl_OntologyAnnotation__OntologyAnnotation_fromJsonString_Static_Z721C83C5", "ARCtrl_OntologyAnnotation__OntologyAnnotation_toJsonString_Static_71136F3F", "ARCtrl_OntologyAnnotation__OntologyAnnotation_ToJsonString_71136F3F", "ARCtrl_OntologyAnnotation__OntologyAnnotation_fromROCrateJsonString_Static_Z721C83C5", "ARCtrl_OntologyAnnotation__OntologyAnnotation_toROCrateJsonString_Static_71136F3F", "ARCtrl_OntologyAnnotation__OntologyAnnotation_ToROCrateJsonString_71136F3F", "ARCtrl_OntologyAnnotation__OntologyAnnotation_fromISAJsonString_Static_Z721C83C5", "ARCtrl_OntologyAnnotation__OntologyAnnotation_toISAJsonString_Static_71136F3F", "ARCtrl_OntologyAnnotation__OntologyAnnotation_ToISAJsonString_71136F3F"]




from typing import Any
from ..fable_library.reflection import (TypeInfo, union_type)
from ..fable_library.types import (Array, Union)

def _expr1860() -> TypeInfo:
    return union_type("ARCtrl.Json.ConverterOptions", [], ConverterOptions, lambda: [[], [], []])


class ConverterOptions(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["ARCtrl", "ROCrate", "ISAJson"]


ConverterOptions_reflection = _expr1860

__all__ = ["ConverterOptions_reflection"]




from abc import abstractmethod
from typing import Any, Generic, Protocol, TypeVar


_T = TypeVar("_T")


class IGenericAdder_1(Protocol, Generic[_T]):
    @abstractmethod
    def Add(self, __arg0: _T, __arg1: _T) -> _T: ...

    @abstractmethod
    def GetZero(self) -> _T: ...


class IGenericAverager_1(Protocol, Generic[_T]):
    @abstractmethod
    def Add(self, __arg0: _T, __arg1: _T) -> _T: ...

    @abstractmethod
    def DivideByInt(self, __arg0: _T, __arg1: int) -> _T: ...

    @abstractmethod
    def GetZero(self) -> _T: ...


class Symbol_wellknown(Protocol):
    @property
    @abstractmethod
    def Symbol_toStringTag(self) -> str: ...


class IJsonSerializable(Protocol):
    @abstractmethod
    def to_json(self) -> Any: ...


SR_indexOutOfBounds: str = "The index was outside the range of elements in the collection."

SR_inputWasEmpty: str = "Collection was empty."

SR_inputMustBeNonNegative: str = "The input must be non-negative."

SR_inputSequenceEmpty: str = "The input sequence was empty."

SR_inputSequenceTooLong: str = "The input sequence contains more than one element."

SR_keyNotFoundAlt: str = "An index satisfying the predicate was not found in the collection."

SR_differentLengths: str = "The collections had different lengths."

SR_notEnoughElements: str = "The input sequence has an insufficient number of elements."

__all__ = [
    "SR_indexOutOfBounds",
    "SR_inputWasEmpty",
    "SR_inputMustBeNonNegative",
    "SR_inputSequenceEmpty",
    "SR_inputSequenceTooLong",
    "SR_keyNotFoundAlt",
    "SR_differentLengths",
    "SR_notEnoughElements",
]
